library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity gc_ipl is
	Port (  rom_adr : out std_logic_vector(20 downto 0);
		rom_dat : inout std_logic_vector(7 downto 0);
		rom_oe : out std_logic;
		rom_we : out std_logic;
		rom_ce : out std_logic;
		rom_reset : out std_logic;
--********************************************************
		-- aus chip sicht!
		uart_clock : in std_logic;
		uart_rx    : in std_logic;
		uart_tx    : out std_logic;
--********************************************************
		-- aus chip sicht!
		exi_in  : in std_logic;
		exi_out : out std_logic;
		exi_clk : in std_logic;
		exi_cs  : in std_logic;
		exi_out_ipl : in std_logic);


end gc_ipl;

architecture behavioral of gc_ipl is

--********************************************************
signal rom_adr_host_sig: STD_LOGIC_VECTOR(20 downto 0);
signal rom_host_oe: std_logic;
signal rom_host_we: std_logic;
signal rom_host_data: std_logic_vector(7 downto 0);

signal exi_count  : std_logic_vector(5 downto 0);
signal exi_data: std_logic_vector(7 downto 0);
signal exi_out_latch: std_logic;
signal exi_loop_through: std_logic;

-- TX stuff: state is: 0 idle, 1 startbit, 2..10 data, 10 stop
signal uart_tx_state: std_logic_vector(3 downto 0);
signal uart_tx_counter: std_logic_vector(3 downto 0);
signal uart_tx_byte: std_logic_vector(7 downto 0);

signal uart_tx_start: std_logic;
signal uart_tx_done: std_logic;

-- RX stuff: state is: 0 idle, 1..9 data, 
signal uart_rx_state: std_logic_vector(3 downto 0);
signal uart_rx_counter: std_logic_vector(3 downto 0);
signal uart_rx_byte: std_logic_vector(7 downto 0);

signal rx_valid, tx_valid: std_logic;

-- protocol stuff:
signal rom_dat_ser: std_logic_vector(7 downto 0);
signal rom_adr_ser: std_logic_vector(19 downto 0);
signal rom_ser_oe, rom_ser_we: std_logic;
signal ser_state  : std_logic_vector(3 downto 0);

-- exi protocol
signal exi_address: std_logic_vector(25 downto 0);
signal config_bootorig: std_logic ;
signal config_bootprefix: std_logic_vector(3 downto 0);
signal do_rom: std_logic;
signal do_config: std_logic;
signal do_write: std_logic;
signal do_next_write: std_logic;
signal do_write_now: std_logic;
signal do_serial: std_logic;


begin



--************************************************************************************************
--  		Global Process
--
-- Test for program or normal mode, and setup signal routes
-- and other signals for it
--************************************************************************************************
Switch_Global: process	(	rom_adr_host_sig,
				rom_dat, uart_clock,
				rom_ser_oe, rom_ser_we, rom_adr_ser, rom_dat_ser, ser_state,
				rom_host_we, rom_host_data, rom_host_oe
			)
begin

--	exi_out <= '0';
	rom_reset <= '1';

--	if ser_state = x"4" then
		rom_ce <= '0';
		rom_adr(20 downto 0) <= (config_bootprefix or rom_adr_host_sig(20 downto 17)) & rom_adr_host_sig(16 downto 0);
		if rom_host_we = '0' then
			rom_dat(7 downto 0) <= rom_host_data;
		else
			rom_dat(7 downto 0) <= (others => 'Z');
		end if;
		rom_oe <= rom_host_oe;
		rom_we <= rom_host_we;
--	else
--		rom_ce <= '0';
--		rom_adr <= "0" & rom_adr_ser;
--		if ser_state /= x"1" and ser_state /= x"6" and ser_state /= x"7" then
--			rom_dat <= (others => 'Z');
--		else
--			rom_dat <= rom_dat_ser;
--		end if;
--		rom_oe <= rom_ser_oe;
--		rom_we <= rom_ser_we;
--	end if;

--	if (uart_clock'event and uart_clock = '1') then
--		if rx_valid = '1' then
--			case uart_rx_byte(7 downto 4) is
--				when x"0" =>
--					rom_dat_ser(3 downto 0) <= uart_rx_byte(3 downto 0);
--				when x"1" =>
--					rom_dat_ser(7 downto 4) <= uart_rx_byte(3 downto 0);
--					ser_state <= x"1";  -- write cycle
--				when x"2" =>
--					case uart_rx_byte(3 downto 0) is
--						when x"0" =>
--							ser_state <= x"2";  -- read cycle w/o postinc
--						when x"1" =>
--							ser_state <= x"3";  -- read cycle with postinc
--						when x"2" =>
--							ser_state <= x"4";  -- exi
--						when others =>
--							ser_state <= x"5";  -- illegal
--					end case;
--				when x"3" =>
--					rom_adr_ser(3 downto 0) <= uart_rx_byte(3 downto 0);
--				when x"4" =>
--					rom_adr_ser(7 downto 4) <= uart_rx_byte(3 downto 0);
--				when x"5" =>
--					rom_adr_ser(11 downto 8) <= uart_rx_byte(3 downto 0);
--				when x"6" =>
--					rom_adr_ser(15 downto 12) <= uart_rx_byte(3 downto 0);
--				when x"7" =>
--					rom_adr_ser(19 downto 16) <= uart_rx_byte(3 downto 0);
--				when others =>
--					ser_state <= x"5";
--			end case;
--		else
--			case ser_state is
--				when x"0" => -- idle
--					rom_ser_we <= '1';
--					rom_ser_oe <= '1';
--					tx_valid <= '0';
--				when x"1" => -- write cycle
--					rom_ser_we <= '0';
--					rom_ser_oe <= '1';
--					ser_state <= x"6";
--				when x"2" => -- read w/o postinc
--					rom_ser_oe <= '0';
--					ser_state <= x"8";
--				when x"3" => -- read w postinc
--					rom_ser_oe <= '0';
--					ser_state <= x"9";
--				when x"6" =>
--					rom_ser_we <= '1';
--					rom_ser_oe <= '1';
--					ser_state <= x"7";
--				when x"7" =>
--					rom_ser_we <= '1';
--					rom_ser_oe <= '1';
--					ser_state <= x"0";
--				when x"8" =>
--					uart_tx_byte <= rom_dat;
--					tx_valid <= '1';
--					ser_state <= x"0";
--				when x"9" =>
--					uart_tx_byte <= rom_dat;
--					tx_valid <= '1';
--					rom_adr_ser(3 downto 0) <= rom_adr_ser(3 downto 0) + '1';
--					ser_state <= x"0";
--				when others =>
--					ser_state <= x"0";
--					uart_tx_byte <= x"40";
--					tx_valid <= '1';
--			end case;
--		end if;
--	end if;
end process;

--UART_TX_process: process (uart_clock, uart_tx_state, uart_tx_start)
--begin
--	if (uart_clock'event and uart_clock = '1') then
--		if uart_tx_state = x"0" then
--			if uart_tx_start = '1' and uart_tx_done = '0' then
--				uart_tx_state <= x"1";
--			end if;
--
--			if uart_tx_start = '0' then
--				uart_tx_done <= '0';
--			end if;
--
--			uart_tx <= '1';
--			uart_tx_counter <= (others => '0');
--		else
--			if (uart_tx_counter = x"0") then
--				if uart_tx_state = x"1" then
--					uart_tx <= '0'; uart_tx_state <= x"2"; -- start
--				elsif (uart_tx_state >= x"2") and (uart_tx_state <= x"A") then
--					if uart_tx_state = x"A" then
--						uart_tx <= '1';
--					else
--						uart_tx <= uart_tx_byte(conv_integer(uart_tx_state) - 2);
--					end if;
--					uart_tx_state <= uart_tx_state + '1';
--				else
--					uart_tx_state <= x"0";
--					uart_tx_done <= '1';
--					uart_tx <= '1';
--				end if;
--			end if;
--			uart_tx_counter <= uart_tx_counter + '1';
--		end if;
--	end if;
--end process;

--UART_RX_process: process (uart_clock, uart_rx_state)
--begin
--	if (uart_clock'event and uart_clock = '1') then
--		if uart_rx_state = x"0" then  -- wait for startbit
--			if uart_rx = '1' then
--				uart_rx_state <= x"0";
--			else
--				uart_rx_state <= x"1";
--			end if;
--			uart_rx_counter <= (others => '0');
--			rx_valid <= '0'; 
--		else
--			if (uart_rx_counter = "1000") then  -- sample in the middle
--				if uart_rx_state >= x"2" and uart_rx_state <= x"9" then
--					uart_rx_byte <= uart_rx & uart_rx_byte(7 downto 1);
--				end if;
--
--				if uart_rx_state = x"A" then  -- stopbit reached?
--					uart_rx_state <= x"0"; -- go into idle
--					rx_valid <= '1';
--				else
--					uart_rx_state <= uart_rx_state + '1';
--				end if;
--			end if;
--			uart_rx_counter <= uart_rx_counter + '1';
--		end if;
--	end if;
--end process;


HOSTReadFlash: process	( exi_cs, exi_clk, rom_dat
			)
	-- 39..8 command phase
	-- 7..0 data...

begin

	if (exi_cs = '1') then                      -- when chip select disabled, just Z
		exi_count <= "100110";
		rom_adr_host_sig <= (others => '0');
		exi_loop_through <= '0';
		exi_out <= 'Z';
		rom_host_oe <= '1';
		rom_host_we <= '1';
		do_rom <= '0';
		do_serial <= '0';
		do_config <= '0';
	else
		if (exi_clk'event and exi_clk = '1') then       -- otherwise sample on raising clock
			              -- one before the first bit of a data, but not the first one
			if ((exi_count(2 downto 0) = "000") and (exi_count(5 downto 4) = "00")) then -- 8 and 0
				      -- increment address and fetch (old) byte
				if do_rom = '1' then
					if exi_address(25) = '0' then
						exi_data <= rom_dat;  -- get OLD byte
						exi_address(10 downto 0) <= exi_address(10 downto 0) + 1;  -- increment address
					end if;
				end if;
			end if;

				-- rom WE enabler
			if do_write_now = '1' and do_rom = '1' then	
				rom_host_we <= '0';
			else
				rom_host_we <= '1';
			end if;

--			if do_write_now = '1' and do_serial = '1' then
--				uart_tx_start <= '1';
--			elsif uart_tx_done = '1' then	-- ack TX done
--				uart_tx_start <= '0';
--			end if;

				-- rom OE handler
			rom_host_oe <= (not do_rom) or (do_write or do_write_now or do_next_write);

			if exi_count = "000110" then
				do_write_now <= do_write;
				do_write <= do_next_write;
				do_next_write <= '0';

				if do_write = '1' and do_config = '1' then
					config_bootorig <= exi_data(1);
					config_bootprefix <= exi_data(7 downto 4);
				end if;
				if do_write = '1' and do_rom = '1' then
					rom_host_data <= exi_data;
				end if;
--				if do_write = '1' and do_serial = '1' and uart_tx_start = '0' then -- only in idle mode
--					uart_tx_byte <= exi_data;
--				end if;

			end if;

			if (exi_count(5 downto 3) = "000") then  -- when not in command phase
				    -- bit counter
				exi_count(2 downto 0) <= exi_count(2 downto 0) - 1;
				    -- output current bit
				if exi_address(25) = '0' then	-- read
					exi_out_latch <= exi_data(conv_integer(exi_count(2 downto 0)));
				else	-- write
					exi_data <= exi_data(6 downto 0) & exi_in;
				end if;
			else
				    -- in command phase, just output ones (or whatever)
				exi_out_latch <= '1';
				exi_count <= exi_count - 1;
				    -- if padding bits are not yet 
				if (conv_integer(exi_count) >= 13) then 
					exi_address <= exi_address(24 downto 0) & exi_in;
				elsif conv_integer(exi_count) = 12 then
					case exi_address(24 downto 21) is
						when x"0" => -- rom access
							exi_loop_through <= config_bootorig;
							do_rom <= '1';
--						when x"5" =>
--							do_serial <= '1';
						when x"6" =>
							do_config <= '1';
						when x"4" => -- sram/rtc access
							exi_loop_through <= '1';
						when others =>
							exi_loop_through <= '1';
					end case;
					do_next_write <= exi_address(25);
				end if;
			end if;
		end if;
		rom_adr_host_sig <= exi_address(20 downto 0);
		if (exi_loop_through = '1') then
			exi_out <= exi_out_ipl;
		else
			exi_out <= exi_out_latch;
		end if;
	end if;	
end process;
--			exi_out <= exi_out_ipl;


end behavioral;



