#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

int fd;

void w(unsigned char c)
{
#if 0
	// startbit
	printf("0 320000 ");
	int i;
	for (i=0; i<8; ++i)
	{
		if (c & (1<<i))
			printf("1 ");
		else
			printf("0 ");
		printf("320000 ");
	}
	printf("1 320000 ");  // stop
	printf("1 1000000 ");  // nen bisschen warten :)
	printf("\n");
#endif
	write(fd, &c, 1);
}

unsigned char r(void)
{
	unsigned char r;
	read(fd, &r, 1);
	return r;
}

int last_addr = -1;

void set_address(int addr)
{
	int diff = last_addr ^ addr;
	if (diff & 0xF)
		w(0x30 | (addr & 0xF));
	if (diff & 0xF0)
		w(0x40 | ((addr>>4) & 0xF));
	if (diff & 0xF00)
		w(0x50 | ((addr>>8) & 0xF));
	if (diff & 0xF000)
		w(0x60 | ((addr>>12) & 0xF));
	if (diff & 0xF0000)
		w(0x70 | ((addr>>16) & 0xF));
	last_addr = addr;
}

unsigned char read8(int addr)
{
	set_address(addr);
	w(0x20);
	return r();
}

unsigned char read8n(int addr)
{
	set_address(addr);
	w(0x21);
			// only the last 4 bits get auto incremented because of cpld space reasons
	last_addr = (last_addr & ~0xF) | ((last_addr + 1) & 0xF);
	return r();
}

void write8(int addr, unsigned char data) 
{
	set_address(addr);
	w(0x00 | (data & 0xF));
	w(0x10 | (data >> 4));
}

void readn(unsigned char *dst, int addr, int len)
{
	while (len--)
		*dst ++ = read8n(addr++);
}

int chip_erase(void)
{
	write8(0xAAA, 0xAA);
	write8(0x555, 0x55);
	write8(0xAAA, 0x80);
	write8(0xAAA, 0xAA);
	write8(0x555, 0x55);
	write8(0xAAA, 0x10);
	printf("erasing..\n");
	while (!(read8(0) & 0x80));
	return 0;
}

int erase_sector(int addr)
{
	write8(0xAAA, 0xAA);
	write8(0x555, 0x55);
	write8(0xAAA, 0x80);
	write8(0xAAA, 0xAA);
	write8(0x555, 0x55);
	write8(addr, 0x30);
	printf("\r%08x erase..\n", addr);
	while (!(read8(addr) & 0x80)) ;
	return 0;
}

int is_eraseblock(int addr)
{
//	if (addr >= 0x10000)
		return !(addr & 0xFFFF);
//	return ((addr == 0) || (addr == 0x4000) || (addr == 0x6000) || (addr == 0x8000));
}

int main(int argc, char **argv)
{
	fd = open("/dev/tts/0", O_RDWR);
	if (fd < 0)
	{
		perror("read");
		return 1;
	}

		// CFI query
	write8(0xAA, 0x98);

	printf("CFI data:\n");	
	int addr;
	for (addr = 0; addr < 0x100; addr += 0x10)
	{
		int i;
		for (i=0; i<0x10; i += 2)
			printf("%02x ", read8n(addr + i));

		for (i=0; i<0x10; i += 2)
		{
			int v = read8(addr + i);
			printf("%c", v >= 0x20 ? v : '.');
		}
		printf("\n");
	}
	
	printf("man id:\n");
	int i;
	for (i=0; i<10; ++i)
	{
		write8(0, 0xF0);
		write8(0xAAA, 0xAA); write8(0x555, 0x55); write8(0xAAA, 0x90);
		printf("%02x ", read8(i * 2));
	}
	printf("\n");
	
	if (argc == 2)
	{
		FILE *fp = fopen(argv[1], "rb");
		
		if (chip_erase())
		{
			printf("chip erase failed!\n");
			return 1;
		}
		
		int addr = 0;
		write8(0xAAA, 0xAA); write8(0x555, 0x55); write8(0xAAA, 0x20);
		while (1)
		{
//			if (is_eraseblock(addr))
//				erase_sector(addr);
			if (!(addr & 0xFF))
			{
				printf("\r%08x", addr);
				fflush(stdout);
			}
			int data = fgetc(fp);
			if (feof(fp))
				break;
			write8(addr, 0xA0);
			write8(addr, data);
			while (read8n(addr) != data);
			++addr;
		}
	} else
	{
		write8(0, 0xf0);
		int len = 1*1024*1024;
		int addr = 0;
		while (len)
		{
			char buffer[1024];
			readn(buffer, addr, 1024);
			write(2, buffer, 1024);
			printf("\r%08x", addr); fflush(stdout);
			len -= 1024;
			addr += 1024;
		}
	}
	return 0;
}
